import sys
import click
import configparser
import os
from os import path
from .service import Service


root = path.realpath(path.join(path.dirname(__file__), '..'))
dconfig_file = path.join(root, 'dconfig.conf')


def read_config():
    config_file = path.join(root, 'config.conf')
    if not path.isfile(config_file):
        if os.name == 'posix':
            config_file = '/etc/gwisp/config.conf'
        else:
            config_file = '%SYSTEMDRIVE%\\gwisp\\config.conf'
    if not path.isfile(config_file):
        raise RuntimeError('Configuration file not found')

    config = configparser.ConfigParser()
    config.read([dconfig_file, config_file])
    app_config = config['app']

    return app_config


def get_service():
    config = read_config()

    return Service(int(config['port']), config['db_url'],
                   config['client_id'], config['client_secret'],
                   config['redirect_uri'], config['jwt_key'])


@click.group()
@click.version_option(version='0.0.1')
def cli():
    pass


@cli.command(help='Register root account')
@click.option('--passwd', required=True, help='Password of root account')
@click.option('--email', required=True, help='Email of root account')
def regroot(passwd, email):
    get_service().reg_root(email, passwd)


@cli.command(help='Empty database')
def renewdb():
    get_service().renew_db()


@cli.command(help='drop database then push sample data to database')
@click.option('--yes', count=True, help='skip confirm message')
def new(yes):
    if not yes:
        confirm = input('This action will drop databse y/n: ')
        if confirm.lower() != 'y':
            sys.exit(0)

    get_service().setup(path.join(root, 'asset/sample-data'))


@cli.command(help='start service')
@click.option('--port', default=9001, help='start gwisp on port')
@click.option('--verbose', count=True, help='show message during run')
def start(port, verbose):
    get_service().start()


@cli.command(help='show information of configuration file')
def info():
    config = read_config()

    print('verbose: {:s}'.format(config['verbose']))
    print('port: {:s}'.format(config['port']))
    print('db_url: {:s}'.format(config['db_url']))
    print('client_id: {:s}'.format(config['client_id']))
    print('client_secret: {:s}'.format(config['client_secret']))
    print('redirect_url: {:s}'.format(config['redirect_url']))



import jsonschema

#: Account identity
attr_id = {
}

#: Password hash
attr_password_hash = {
    'type': 'string'
}

#: Created date
attr_created_date = {
    'type': 'string',
    'format': 'date-time'
}

#: Modified date
attr_modified_date = {
    'type': 'string',
    'format': 'date-time'
}

#: Schema of item use to insert
_schema_insert = {
    'type': 'object',
    'required': ['_id', 'password_hash', 'created_date'],
    'properties': {
        '_id': attr_id,
        'password_hash': attr_password_hash,
        'created_date': attr_created_date
    }
}

#: Schema of item use to update
_schema_update = {
    'type': 'object',
    'required': ['_id'],
    'properties': {
        '_id': attr_id,
        'password_hash': attr_password_hash,
        'modified_date': attr_modified_date
    }
}


def validate_insert(item):
    '''
    Validate password and meta data

    :param dict item: Password and meta data
    '''

    jsonschema.validate(item, _schema_insert)


def validate_update(item):
    '''
    Update password and meta data

    :param dict item: Password object, contains '_id' fields
    '''

    jsonschema.validate(item, _schema_update)



import jsonschema


#: Identify
attr_id = {
}

#: Name
attr_name = {
    'type': 'string',
    'pattern': '^[a-z0-9 -.]{1,64}$'
}

#: Author
attr_author = {
}

#: Tags
attr_tags = {
    'type': 'array',
    'items': {
        'type': 'string',
        'pattern': '^[a-z0-9]{1,16}$'
    }
}

#: Tasks
attr_tasks = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'start': {
                'type': 'string',
                'pattern': '^([0-1][0-9]|2[0-3]):[0-5][0-9]$'
            },
            'action': {
                'type': 'string',
                'pattern': '^[a-zA-Z0-9 ]{1,32}$'
            }
        }
    }
}

#: Notes
attr_notes = {
    'type': 'string',
    'pattern': '^(.|\n){0,1024}$'
}

#: Item use to insert
_schema_insert = {
    'required': ['name', 'author', 'tags'],
    'type': 'object',
    'properties': {
        'name': attr_name,
        'author': attr_author,
        'tags': attr_tags,
        'tasks': attr_tasks,
        'notes': attr_notes
    }
}

#: Item use to update
_schema_update = {
    'required': ['_id'],
    'type': 'object',
    'properties': {
        '_id': attr_id,
        'name': attr_name,
        'tags': attr_tags,
        'tasks': attr_tasks,
        'notes': attr_notes
    }
}


def validate_insert(items):
    '''
    Validate scheduler use to insert

    :param object items: List or single of schedulers
    '''

    if type(items) is list:
        for item in items:
            jsonschema.validate(item, _schema_insert)
    else:
        jsonschema.validate(items, _schema_insert)


def validate_update(items):
    '''
    Validate scheduler to update

    :param object items: List or single of schedulers
    '''

    if type(items) is list:
        for item in items:
            jsonschema.validate(items, _schema_update)
    else:
        jsonschema.validate(items, _schema_update)



import jsonschema

#: Root account
attr_root_name = {
    'type': 'string',
    'regex': '^[a-zA-Z0-9 -.]{1,64}$'
}

#: Installed date
attr_date = {
    'type': 'string',
    'format': 'date-time'
}

#: Internet address installed
attr_ip = {
    'type': 'string',
    'format': 'ipv4',
}

#: Item to insert
_schema_insert = {
    'type': 'object',
    'required': ['root_name', 'date', 'ip'],
    'properties': {
        'root_name': attr_root_name,
        'date': attr_date,
        'ip': attr_ip
    }
}


def validate_insert(item):
    '''
    Validate setting to insert

    :param dict item: Item to validate
    '''

    jsonschema.validate(item, _schema_insert)



from gunicorn.app.base import BaseApplication

from .worker import Worker


class Server(BaseApplication):
    '''
    Use gunicorn to make server. It will be create worker, auto restart
    after worker terminate

    :param int port: Port which is service will be serve
    :param str db_url: Url to mongo database server
    :param str client_id: Google client identity, use for oauth-2
    :param str client_secret: Google client secret, use for oauth-2
    :param str redirect_uri: Callback url after confirm google login
    :param str jwt_key: Secret words use to sign token of service
    '''

    def __init__(self, port, db_url,
                 client_id, client_secret, redirect_uri, jwt_key):

        self._port = port
        self._db_url = db_url
        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        self._jwt_key = jwt_key

        super(Server, self).__init__()

    def load_config(self):
        '''
        Override method of BaseApplication of gunicorn
        '''

        if self._port > 0:
            self.cfg.set('bind', '0.0.0.0:{}'.format(self._port))

    def load(self):
        '''
        Override method of BaseApplication of gunicorn
        '''

        worker = Worker(self._db_url,
                        self._client_id, self._client_secret,
                        self._redirect_uri, self._jwt_key)
        return worker.start()



from ..model.account import validate_insert, validate_update,\
                            validate_groups
from ..error import NotExistError, NotAllowError


verify_fields = {
    '_id': 0,
    'subject': 0,
    'name': 0,
    'groups': 0,
    'scheduler': 0,
    'schedulers': 0,
    'language': 0
}


class AccountRepo(object):
    '''
    Provide method to query account information in database

    :param pymongo.MongoClient db: Instance of connection to mongodb server
    '''
    def __init__(self, db):
        self._db = db
        self._coll = db['account']
        self._sched_coll = db['scheduler']

    def find_by_id(self, id):
        '''
        Find on account by identity

        :param bson.objectid.ObjectId id: Identity of account
        :return: Account dictionary on found
        :return: None on not found
        :rtype: dict
        '''

        return self._coll.find_one({'_id': id})

    def find_by_name(self, name):
        '''
        Find account match with name

        :param str name: Account name
        '''

        return self._coll.find_one({'name': name})

    def find_by_sub(self, sub):
        '''
        Find account match with subject

        :param str sub: Subject to find
        :return: Account directory on found
        :return: None on not found
        :rtype: dict
        '''
        return self._coll.find_one({'subject': sub})

    def insert_one(self, item):
        '''
        Insert one account into database

        :param dict item: Account dictionary
        :return: Identity of new account
        :rtype: bson.objectid.ObjectId
        '''

        if 'groups' not in item:
            item['groups'] = []
        if 'schedulers' not in item:
            item['schedulers'] = []

        validate_insert(item)

        return self._coll.insert_one(item).inserted_id

    def is_exist(self, ids):
        '''
        Verify one account is exist

        :param bson.objectid.ObjectId id: Account identity
        :return: True on account is exist
        :return: False on account is not exist
        :rtype: bool
        '''

        return self._coll.find_one({'_id': ids}, verify_fields) is not None

    def is_exist_list(self, ids):
        '''
        Verify list of account identity is exist

        :param list[bson.objectid.ObjectId] ids: List of account identity
        :return: True on all of account is exist
        :return: False on one or more account is not exist
        :rtype: bool
        '''
        cursor = self._coll.find({'_id': {'$in': ids}}, verify_fields)
        return cursor.count() == len(ids)

    def update_one(self, item):
        '''
        Update one account. Exclude name and subject fields. Item must specify
        _id field to detect what is account to update. Function only update
        fields is specify in item, other fields will be skip

        :param dict item: Account dictionary to update
        '''

        if 'email' in item:
            raise NotAllowError('Not allow update email of account')
        if 'name' in item:
            raise NotAllowError('Not allow update name of account')
        if 'subject' in item:
            raise NotAllowError('Not allow update subject of account')

        validate_update(item)
        id = item['_id']
        del item['_id']

        self._coll.update_one({'_id': id}, {'$set': item})

    def use_scheduler(self, acc_id, sched_id):
        '''
        Mark account use specify scheduler and append scheduler into
        schedulers list

        :param bson.objectid.ObjectId acc_id: Account identity which is
            use scheduler
        :param bson.objectid.ObjectId sched_id: Scheduler identity which is
            used by account
        '''

        # scheduler must be exist
        if self._sched_coll.find_one({'_id': sched_id}) is None:
            raise NotExistError('scheduler is not exist')

        # make scheduler is using
        # push scheduler to schedulers of account
        patch = {
            '$set': {'scheduler': sched_id},
            '$addToSet': {'schedulers': sched_id}
        }

        self._coll.update_one({'_id': acc_id}, patch)

    def remove_scheduler(self, acc_id, sched_id):
        '''
        Remove scheduler from schedulers list

        :param bson.objectid.ObjectId acc_id: Account identity which will
            remove scheduler
        :param bson.objectid.ObjectId sched_id: Scheduler identity will
            be remove
        '''

        patch = {
            '$pull': {'schedulers': sched_id}
        }
        self._coll.update_one({'_id': acc_id}, patch)

    def add_to_groups(self, acc_id, groups):
        '''
        Add account to groups

        :param bson.objectid.ObjectId acc_id: Account identity will be
            add to groups
        :param str group: Groups which account will be add to
        '''

        validate_groups(groups)

        patch = {
            '$addToSet': {'groups': {'$each': groups}}
        }

        self._coll.update_one({'_id': acc_id}, patch)

    def remove_from_groups(self, acc_id, groups):
        '''
        Remove account from groups

        :param bson.objectid.ObjectId acc_id: Account identity will be
            remove from groups
        :param str group: Groups which account will be remove from
        '''

        validate_groups(groups)

        patch = {
            '$pull': {'groups': {'$in': groups}}
        }

        self._coll.update_one({'_id': acc_id}, patch)



from ..error import AuthenticateError


class BasicLoginCtl(object):
    '''
    Login by username and password

    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, protector):
        self._protector = protector

    def on_get(self, req, res):
        '''
        Request: Username and password in header

        Response: Token string
        '''

        if 'USERNAME' not in req.headers.keys():
            raise AuthenticateError('Username header is not set')
        if 'PASSWORD' not in req.headers.keys():
            raise AuthenticateError('Passowrd header is not set')

        token = self._protector.login_by_passwd(
            req.headers['USERNAME'],
            req.headers['PASSWORD']
        )

        res.body = {
            'access_token': token
        }


class LoginUrlCtl(object):
    '''
    Information about how to login

    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, protector):
        self._protector = protector

    def on_get(self, req, res):
        '''
        Response: Login url
        '''

        res.body = {
            'url': self._protector.login_url
        }


class LoginCtl(object):
    '''
    Support login by google oauth-2

    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, protector):
        self._protector = protector

    def on_get(self, req, res):
        '''
        Request: Login code from google as query param

        Response: Token string
        '''

        access_token = self._protector.login_by_code(req.params['code'])
        res.body = {
            'access_token': access_token
        }


class MeCtl(object):
    '''
    Account information

    :param pymongo.MongoClient db: Instance of mongo client
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._db = db
        self._protector = protector

    def on_get(self, req, res):
        '''
        Request: Token in header

        Response: Account information
        '''

        info = self._protector.check(req, True)
        res.body = info.account



import httplib2
import hashlib

import jwt
import datetime
from bson.objectid import ObjectId
from oauth2client.client import OAuth2WebServerFlow
from apiclient import discovery

from .error import AuthenticateError, AuthorizeError
from .repo import PasswordRepo, AccountRepo
from .util import AuthInfo


class Protector(object):
    '''
    Control authenticating and authorizing

    :param pymongo.MongoClient db: Instance of connection to mongo server
    :param string client_id: Google oauth-2 client identify
    :param string client_secret: Google oauth-2 client secret
    :param list scope: Google oauth-2 scope
    :param string redirect_uri: Oauth-2 callback uri
    :param string jwt_key: Secret word use to sign token
    '''
    def __init__(self, db, client_id, client_secret, scope, redirect_uri,
                 jwt_key):
        self._db = db
        self._client_id = client_id
        self._client_secret = client_secret
        self._scope = scope
        self._redirect_uri = redirect_uri
        self._jwt_key = jwt_key

        self._acc_repo = AccountRepo(db)
        self._passwd_repo = PasswordRepo(db)

    def create_account(self, account, password):
        '''
        Create an account with password

        :param dict account: Account information
        :param str password: Password to login
        '''

        acc_id = self._acc_repo.insert_one(account)

        passwd_hash = hashlib.sha224(password.encode('utf-8')).hexdigest()
        passwd = {
            '_id': acc_id,
            'password_hash': passwd_hash
        }

        self._passwd_repo.insert_one(passwd)

    @property
    def login_url(self):
        '''
        Get oauth-2 login url. Use browser and url to login by google oauth-2
        service

        :return: Url use to login
        :rtype: str
        '''

        oauth = self._gen_oauth2()
        return oauth.step1_get_authorize_url()

    def login_by_passwd(self, name, password):
        '''
        Login by account name and password

        :param str name: Name of account
        :param str password: Password
        :return: Token string
        :rtype: str
        '''

        account = self._acc_repo.find_by_name(name)
        if account is None:
            raise AuthenticateError('Account is not exist')

        passwd_hash = hashlib.sha224(password.encode('utf-8')).hexdigest()
        match = self._passwd_repo.is_exist(account['_id'], passwd_hash)
        if match is False:
            raise AuthenticateError('Password is incorrect')

        return self._gen_token(account['_id'])

    def login_by_code(self, code):
        '''
        Login by code from google. If google account has not early login,
        create an account

        :param str code: Code what is received from google oauth-2 service
        :return: Token what is to authenticate and authorize
        :rtype: string
        '''

        # exchange code to tokens
        oauth = self._gen_oauth2()
        credentials = oauth.step2_exchange(code)

        # use token to get google account information
        http = credentials.authorize(httplib2.Http())
        service = discovery.build('plus', 'v1')
        req = service.people().get(userId='me')
        me = req.execute(http=http)

        # verify google account is early register in system
        # if not, register first
        sub = credentials.id_token['sub']
        acc = self._acc_repo.find_by_sub(sub)
        if acc is None:
            acc = {
                'subject': sub,
                'email': credentials.id_token['email'],
                'name': me['displayName'],
                'language': me['language'],
                'avatar': me['image']['url']
            }
            self._acc_repo.insert_one(acc)

        # system token
        return self._gen_token(acc['_id'])

    def check(self, req, iacc=False, group=None):
        '''
        Authenticate and authorize if **group** parameter is specify

        :param falcon.Request req: Contains information about http request
        :param boolean iacc: Decide include account information
        :param string group: Group what is expected account in
        :return: Information after authenticate and authorize
        :rtype: gwisp.AuthInfo
        '''

        token = self.authenticate(req.headers['AUTHORIZATION'])
        acc_id = ObjectId(token['sub'])
        acc = None

        if iacc is True:
            acc = self._acc_repo.find_by_id(acc_id)
            if acc is None:
                raise AuthenticateError('Can not identity user')

        if group is not None:
            if iacc is False:
                acc = self._acc_repo.find_by_id(acc_id)
            if acc is None:
                raise AuthenticateError('Can no identity user')
            if group not in acc['group']:
                raise AuthorizeError('Invalid permission')

        return AuthInfo(acc_id, token, acc)

    def authenticate(self, token):
        '''
        Identify account by decode token

        :param string token: Token to decode
        :return: Token has decoded
        :rtype: dict
        '''

        return jwt.decode(token, self._jwt_key)

    def _gen_oauth2(self):
        '''
        Create an oauth-2 web server flow object

        :return: Instance of OAuth2WebServerFlow
        :rtype: oauth2client.client.OAuth2WebServerFlow
        '''

        return OAuth2WebServerFlow(
            client_id=self._client_id,
            client_secret=self._client_secret,
            scope=self._scope,
            redirect_uri=self._redirect_uri)

    def _gen_token(self, acc_id):
        '''
        Create token string

        :param bson.objectid.ObjectId acc_id: Identity of account
        :return: Token string
        :rtype: str
        '''

        token_data = {
            'sub': str(acc_id),
            'exp': datetime.datetime.now().timestamp() + 3600
        }
        token = jwt.encode(token_data,
                           self._jwt_key,
                           algorithm='HS512')

        return token.decode()



from datetime import datetime

from ..model.password import validate_insert, validate_update
from ..error import NotAllowError


class PasswordRepo(object):
    '''
    Password repository
    Use to store password hashed and meta data related with password

    :param pymongo.MongoClient db: Instance of connection to mongodb server
    '''

    def __init__(self, db):
        self._db = db
        self._coll = db['password']

    def is_exist(self, id, passwd_hash):
        '''
        Check pair identity-password is exist

        :param bson.objectid.ObjectId id: Identity of account
        :param str password: Password
        '''

        filters = {'_id': id, 'password_hash': passwd_hash}

        return self._coll.find_one(filters) is not None

    def insert_one(self, item):
        '''
        Insert one password object

        :param dict item: Password hashed and meta data
        '''

        if 'created_date' in item:
            raise NotAllowError('created_date is auto fill')
        if 'modified_date' in item:
            raise NotAllowError('modified_date is auto fill')

        item['created_date'] = datetime.now().isoformat()
        item['modified_date'] = datetime.now().isoformat()

        validate_insert(item)

        self._coll.insert_one(item)

    def update_one(self, item):
        '''
        Update password

        :param dict item: New password hashed
        '''

        if 'created_date' in item:
            raise NotAllowError('created_date is not allowed to update')
        if 'modified_date' in item:
            raise NotAllowError('modified_date is auto fill')

        item['modified_date'] = datetime.now().isoformat()

        validate_update(item)

        id = item['_id']
        del item['_id']

        self._coll.update_one({'_id': id}, {'$set': item})



class RootCtl(object):
    '''
    Information of service. Read only resource
    '''

    def __init__(self):
        pass

    def on_get(self, req, res):
        '''
        Response: Essential information of service
        '''

        data = {
            'name': 'gwisp service',
            'notes': 'http programing interface of scheduler storage',
            'version': '0.0.1',
            'source': 'https://github.com/kevin-leptons/gwisp',
            'pip': 'https://pypi.python.org/pypi/gwisp/0.0.4',
            'document_url': 'http://gwisp.readthedocs.io/en/latest/'
        }

        res.body = data



from falcon import HTTP_204, HTTPBadRequest
from bson.objectid import ObjectId
from ..repo import SchedulerRepo
from ..util import parse_selector


class SchedulerCtl(object):
    '''
    Scheduler resource

    :param pymongo.MongoClient db: Instance of mongo client
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._protector = protector
        self._repo = SchedulerRepo(db)

    def on_get(self, req, res):
        '''
        Search scheduler

        Request: Query parameters as selector

        Response: List of scheduler match with selector
        '''

        res.body = self._repo.search(parse_selector(req))

    def on_post(self, req, res):
        '''
        Create new scheduler

        Request: Scheduler in body, access token in header

        Response: Relative location of new scheduler in header, field
            Location
        '''

        if req.context['body'] is None:
            raise HTTPBadRequest('Invalid body', 'Body is empty')

        auth = self._protector.check(req)

        item = req.context['body']
        item['author'] = auth.account_id
        id = self._repo.insert_one(item)

        res.status = HTTP_204
        res.set_header('Location', '/scheduler/{}'.format(str(id)))


class SchedulerItemCtl(object):
    '''
    Scheduler resource

    :param pymongo.MongoClient db: Instance of mongo client
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._repo = SchedulerRepo(db)
        self._protector = protector

    def on_get(self, req, res, id):
        '''
        Query single scheduler

        Request: Identify of scheduler in request url

        Response: Scheduler in body
        '''

        res.body = self._repo.find_by_id(ObjectId(id))

    def on_delete(self, req, res, id):
        '''
        Remove scheduler

        Request: Identify of scheduler in request url
        '''

        self._protector.check(req, False, 'root')



import re
from ..model.scheduler import validate_insert, validate_update
from ..error import NotExistError
from ..util import Selector


search_fields = {'notes': 0, 'tasks': 0, 'tags': 0}
verify_fields = {
    '_id': 0,
    'name': 0,
    'tags': 0,
    'tasks': 0,
    'notes': 0
}


class SchedulerRepo(object):
    '''
    Provide method to query scheduler in mongodb

    :param pymongo.MongoClient db: Instance of MongoClient
    '''
    def __init__(self, db):
        self._db = db
        self._coll = db['scheduler']
        self._acc_coll = db['account']

    def find_by_id(self, id):
        '''
        Find scheduler by identity

        :param bson.objectid.ObjectId id: Identity of scheduler
        :return: Scheduler dictionary on found
        :return: None of not found
        :rtype: dict
        '''
        return self._coll.find_one({'_id': id})

    def search(self, selector=Selector()):
        '''
        Search scheduler match with selector

        :param gwisp.Selector selector: Contains specification to search
        :return: List of dictionary contains essential information of scheduler
              on found
        :return: List with none element not found
        :rtype: list
        '''

        filters = {}
        if selector.keyword is not None:
            filters = self._gen_filter(selector.keyword)

        cursor = self._coll.find(filters,
                                 search_fields,
                                 skip=selector.skip,
                                 limit=selector.page_size)
        return list(cursor)

    def search_by_author(self, author_id, selector=Selector()):
        '''
        Search by author identity and match with selector

        :param bson.objectid.ObjectId author_id: Author identity
        :param gwisp.Selector selector: Contains specification to search
        :return: List of dictionary contains essential information of scheduler
              on found
        :return: List with none element not found
        :rtype: list
        '''

        filters = {'$and': [{'author': author_id}]}
        if selector.keyword is not None:
            filters['$and'].append(self._gen_filter(selector.keyword))

        cursor = self._coll.find(filters,
                                 search_fields,
                                 skip=selector.skip,
                                 limit=selector.page_size)

        return list(cursor)

    def search_by_id(self, ids):
        '''
        Search by list of identity of schedulers

        :param list ids: List of identity of schedulers
        :return: List of dictionary contains essential information of scheduler
              on found
        :return: List with none element not found
        :rtype: list
        '''

        filters = {'_id': {'$in': ids}}

        cursor = self._coll.find(filters, search_fields)

        return list(cursor)

    def insert_one(self, item):
        '''
        Insert one scheduler

        :param dict item: Scheduler to insert
        :return: Identity of new scheduler
        :rtype: bson.objectid.ObjectId
        '''

        validate_insert(item)

        if (not self._acc_is_exist(item['author'])):
            raise NotExistError('account is not exists')

        return self._coll.insert_one(item).inserted_id

    def insert_many(self, items):
        '''
        Insert many schedulers

        :param list items: List of schedulers
        :return: List of identity of new scheduler
        :rtype: list
        '''

        validate_insert(items)

        author_ids = []
        for item in items:
            if item['author'] not in author_ids:
                author_ids.append(item['author'])
        if (not self._acc_is_exist(author_ids)):
            raise NotExistError('account is not exists')

        return self._coll.insert_many(items).inserted_ids

    def update_one(self, item):
        '''
        Update one scheduler

        :param dict item: Scheduler to update, must contain _id fields
        '''

        validate_update(item)
        id = item['_id']
        del item['_id']

        return self._coll.update_one({'_id': id}, {'$set': item})

    def delete_one(self, id):
        '''
        Delete one scheduler

        :param bson.objectid.ObjectId id: Identity of scheduler to delete
        '''

        return self._coll.delete_one({'_id': id})

    def is_exist(self, id):
        '''
        Verify that scheduler is exist

        :param id: Identity of scheduler
        :type id: bson.objectid.ObjectId
        :return: True on scheduler is exist
        :return: False on scheduler is not exist
        :rtype: bool
        '''

        return self._coll.find_one({'_id': id}, verify_fields) is not None

    def _acc_is_exist(self, ids):
        '''
        Verify list of account is exist

        :param ids: List of account identity
        :type ids: list
        :return: True on all of account is exist
        :return: False on one or more account is not exist
        :rtype: bool
        '''

        if type(ids) is list:
            cursor = self._acc_coll.find({'_id': {'$in': ids}})
            return cursor.count() == len(ids)
        else:
            return self._acc_coll.find_one({'_id': ids}) is not None

    def _gen_filter(self, keyword):
        '''
        Generate query filter of mongodb

        :param str keyword: Keyword to search
        :return: Filter
        :rtype: dict
        '''
        return {'$or': [
            {'name': {'$regex': re.compile(keyword)}},
            {'tags': {
                '$elemMatch': {
                    '$regex': re.compile(keyword)
                }
            }}
        ]}



import jsonschema
from .scheduler import attr_id as sched_id, attr_name as sched_name,\
                       attr_tags as sched_tags, attr_tasks as sched_tasks,\
                       attr_notes as sched_notes


#: Identify
attr_id = {

}

#: Email
attr_email = {
    'type': 'string',
    'pattern': '^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$'
}

#: Name
attr_name = {
    'type': 'string',
    'pattern': '^[a-z0-9 -.]{1,64}$'
}

#: Subject
attr_subject = {
    'type': 'string'
}

#: Groups
attr_groups = {
    'type': 'array',
    'items': {
        'type': 'string',
        'pattern': '^[a-z0-9]{1,8}$'
    }
}

#: Scheduler
attr_scheduler = {
    'type': 'object',
    'properties': {
        '_id': sched_id,
        'name': sched_name,
        'tags': sched_tags,
        'tasks': sched_tasks,
        'notes': sched_notes
    }
}

#: Schedulers
attr_schedulers = {
    'type': 'array',
    'items': sched_id
}

#: Language
attr_language = {
    'type': 'string',
    'pattern': '^[a-z]{2}$'
}

#: Item use to insert
_schema_insert = {
    'type': 'object',
    'required': ['email', 'name', 'subject', 'language'],
    'properties': {
        'email': attr_email,
        'name': attr_name,
        'subject': attr_subject,
        'groups': attr_groups,
        'scheduler': attr_scheduler,
        'schedulers': attr_schedulers,
        'language': attr_language
    }
}

#: Item use to update
_schema_update = {
    'type': 'object',
    'required': ['_id'],
    'properties': {
        '_id': attr_id,
        'groups': attr_groups,
        'scheduler': attr_scheduler,
        'schedulers': attr_schedulers,
        'language': attr_language
    }
}


def validate_insert(item):
    '''
    Validate account will be create

    :param object item: Account object
    '''

    jsonschema.validate(item, _schema_insert)


def validate_update(item):
    '''
    Validate account will be update

    :param object item: Account object
    '''

    jsonschema.validate(item, _schema_update)


def validate_groups(groups):
    '''
    Validate groups of account

    :param object item: Account object
    '''

    jsonschema.validate(groups, attr_groups)
